#include "port_mapping.h"

#include "design.h"
#include "hierarchical_channel_instance.h"
#include "instance.h"
#include "module.h"
#include "port.h"
#include "primitive_channel_instance.h"

std::list<risc::InstanceTree>
risc::get_mapped_channel_instances(Port *port)
{
  // This list stores all channels which are mapped to the given port.
  // The port can occur in many instance trees, so, one port can be bound
  // to many channels
  std::list<InstanceTree> mapped_channels;

  // here we are searching for all instance trees which include this port
  InstanceTree root = risc::get_design()->get_root();
  std::list<InstanceTree> port_instances
    = root.find_all_instances_of(port->get_ast_node());

  // Iteration over all port instance trees
  for(std::list<InstanceTree>::iterator
      iter  = port_instances.begin();
      iter != port_instances.end();
      iter++) {

    mapped_channels.push_back(get_mapped_channel_instance(*iter));
  }

  return mapped_channels;
}

SgVariableSymbol*
risc::get_mapped_symbol(
    SgVariableSymbol *port,
    Module *parent_module,
    SgVariableDefinition *parent_variable)
{
  SgVariableSymbol *result = NULL;

  // We are searching for the port binding only in the function
  // 'before_end_of_elaboration'.
  SgFunctionDefinition *before_end_of_elaboration
    = parent_module->get_before_end_of_elaboration_definition();

  // Iterating over all statements in the function 'before_end_of_elaboration'
  for(SgStatementPtrList::iterator
      iter  = before_end_of_elaboration->get_body()->get_statements().begin();
      iter != before_end_of_elaboration->get_body()->get_statements().end();
      iter++) {

    SgExprStatement *expr_stmt = isSgExprStatement(*iter);

    SgFunctionCallExp *func_call
      = isSgFunctionCallExp(expr_stmt->get_expression());

    // First check if the statement contains a function call
    if(func_call) {

      // TODO: quick hack, I should look for the symbol
      if(SageInterface::get_name(func_call->getAssociatedFunctionDeclaration())
          == "bind") {

        // Please look in a corresponding AST for the documentation
        // given: asdf.jkl.bind(var);
        // var_ref_exp_matching = asdf
        // var_ref = jkl

        SgExpression *function = func_call->get_function();

        SgDotExp *dot_expr = isSgDotExp(function);
        SgCastExp *cast_exp = isSgCastExp(dot_expr->get_lhs_operand());

        SgDotExp *dot_expr2 = NULL;
        SgVarRefExp *var_ref = NULL;
        SgVarRefExp *var_ref_exp_matching = NULL;

        if(cast_exp) {
          dot_expr2 = isSgDotExp(cast_exp->get_operand());

          // work around for sc_out
          if(dot_expr2 == NULL) {

            dot_expr2
              = isSgDotExp(isSgCastExp(isSgCastExp(cast_exp->get_operand())
                    ->get_operand())->get_operand());
          }

          var_ref = isSgVarRefExp(dot_expr2->get_rhs_operand());
          SgArrowExp *arrow_exp = isSgArrowExp(dot_expr2->get_lhs_operand());
          var_ref_exp_matching = isSgVarRefExp(arrow_exp->get_rhs_operand());

        } else {
          // this works for the bind of sc_in
          dot_expr2 = isSgDotExp(dot_expr->get_lhs_operand());

          var_ref_exp_matching
            = isSgVarRefExp(isSgArrowExp(dot_expr2->get_lhs_operand())
                ->get_rhs_operand());

          var_ref = isSgVarRefExp(dot_expr2->get_rhs_operand());
        }

        if(var_ref->get_symbol() == port) {

          // Here we are checking, if the parent variable is matching the
          // binding variable.
          if(var_ref_exp_matching->get_symbol() !=
              parent_variable->get_vardefn()->
                search_for_symbol_from_symbol_table()) {
            continue;
          }

          // is binded to this one
          std::vector<SgNode*> binded_element
            = NodeQuery::querySubTree(func_call->get_args(), V_SgVarRefExp);

          return isSgVarRefExp(binded_element.at(0))->get_symbol();
        }
      }
    }
  }

  return result;
}


risc::InstanceTree
risc::get_mapped_channel_instance(InstanceTree port)
{
  SgVariableSymbol *port_symbol = port.get_object()->get_symbol();

  // going one step up in the hierarchy
  port.path_.pop_back();

  // getting the mapped element in the parent class
  SgVariableSymbol *mapped_object = get_mapped_symbol(
      port_symbol,
      dynamic_cast<Module*>(port.get_scope()),
      port.get_object()->get_ast_node());

  // here we are checking the connection is a port
  Port *intermediate_port = port.get_scope()->find_port(
      isSgVariableDefinition(mapped_object->get_declaration()
        ->get_definition()));

  // here we are going in the hierarchy up until the is no longer a port to
  // port mapping
  while(intermediate_port) {

    port_symbol = intermediate_port->get_symbol();

    // going one step up in the hierarchy
    port.path_.pop_back();

    // getting the mapped element in the parent class
    mapped_object = get_mapped_symbol(
                        port_symbol,
                        dynamic_cast<Module*>(port.get_scope()),
                        port.get_object()->get_ast_node());

    // here we are checking the connection is a port
    intermediate_port = port.get_scope()->find_port(
      isSgVariableDefinition(mapped_object->get_declaration()
        ->get_definition()));
  }

  // Finally, we have to create the instance tree for the mapped channel
  // In the following, we are sorting out if it is a primitive or a
  // hierarchical channel
  SgVariableDefinition *channel_var_def =
    isSgVariableDefinition(mapped_object->get_declaration()->get_definition());

  Instance *mapped_channel
    = port.get_scope()->find_prim_channel(channel_var_def);

  if(mapped_channel) {

    port.path_.back().second = mapped_channel;
  } else {

    port.path_.back().second
      = port.get_scope()->find_hier_channel(channel_var_def);
  }

  return port;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
